#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

from typing import  List
from collections import OrderedDict
from copy import deepcopy

#from keysql_client.keysql_store import KeySqlStore
#from keysql_client.keysql_catalog_statement import *
#from keysql_client.key_object_statement import CreateKeyObjectStatement,DropKeyObjectStatement,AlterKeyObjectStatement

#from keysql_client.key_object import *

from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage
import logging

from keysql_client.keysql_instance import KeySql_Instance
from keysql_client.keysql_parser import KeySqlParser
from keysql_client.key_object import KeyObjectChar, KeyObjectNumber, KeyObjectDate, KeyObjectInteger, KeyObjectMultiple, KeyObjectComposition
from keysql_client.keysql_catalog import KeySqlCatalog
from keysql_client.keysql_store import KeySqlStore

class KeySqlClassFromInstance:

    @staticmethod
    def create_keyobject_from_instance(ko_def, ko_classes=[]):
        # example keysql format (str)
        # {KEYOBJECT_DESCRIPTION:{CATALOG_NAME:'CATALOG_DUMMY',
        # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'NUMBER1',KEYOBJECT_TYPE:'NUMBER',COMPONENT_KEYOBJECT_NAMES:NULL},COMPONENT_KEYOBJECTS:NULL}}

        # example json format (dict)
        # {'KEYOBJECT_DESCRIPTION': {'CATALOG_NAME': 'CATALOG_DUMMY',
        # 'KEYOBJECT_DEFINITION': {'KEYOBJECT_NAME': 'NUMBER1', 'KEYOBJECT_TYPE': 'NUMBER', 'COMPONENT_KEYOBJECT_NAMES': None}, 'COMPONENT_KEYOBJECTS': None}}

        if type(ko_def) is dict:
            #print(ko_def, '--------------ko_def dict')
            if 'KEYOBJECT_INFO'.lower() in ko_def:
                ko = ko_def['KEYOBJECT_INFO'.lower()]   # 'KEYOBJECT_DESCRIPTION'
            else:
                ko = ko_def
            ko = ko_def['KEYOBJECT_DEFINITION'.lower()]
        elif type(ko_def) is KeySql_Instance and ko_def.instance_name() == 'KEYOBJECT_DEFINITION':   # not from show catalog, but from show keyobjects)
            ko = ko_def.instance_value()
        elif type(ko_def) is type:      # TODO: for ad-hoc keyobjects
            print(ko_def, '------------------x1')
            print(ko_def.__dict__, '-------------x2')
            return KeySqlClassFromInstance.create_keyobject_from_instance(ko_def.__dict__, ko_classes=ko_classes)
        else:
            #print("type(ko_def):", type(ko_def), '-----------------------------not implemented1')
            #print(ko_def, '-----------------------------k0_def not implemented1')
            raise Exception('NOT IPLEMENTED YET (KeySqlClassFromInstance.create_keyobject_from_instance)')

        if ko['KEYOBJECT_TYPE'.lower()] == 'NUMBER':
            return KeyObjectNumber(ko['KEYOBJECT_NAME'.lower()])

        # {'KEYOBJECT_DEFINITION': {'KEYOBJECT_NAME': 'KO2', 'KEYOBJECT_TYPE': 'CHAR', 'COMPONENT_KEYOBJECT_NAMES': None}}
        if ko['KEYOBJECT_TYPE'.lower()] == 'CHAR':
            return KeyObjectChar(ko['KEYOBJECT_NAME'.lower()])

        if ko['KEYOBJECT_TYPE'.lower()] == 'DATE':
            return KeyObjectDate(ko['KEYOBJECT_NAME'.lower()])

        if ko['KEYOBJECT_TYPE'.lower()] == 'INTEGER':
            return KeyObjectInteger(ko['KEYOBJECT_NAME'.lower()])

        # {KEYOBJECT_DESCRIPTION:{CATALOG_NAME:'CATALOG_DUMMY',KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'NUMBERS',KEYOBJECT_TYPE:'MULTIPLE',
        # COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'NUMBER1'}},
        # COMPONENT_KEYOBJECTS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'NUMBER1',KEYOBJECT_TYPE:'NUMBER',COMPONENT_KEYOBJECT_NAMES:NULL}}}}
        """
        print(ko, '_______________________------------------------------ko')
        print(ko['KEYOBJECT_NAME'].instance_name())
        print(ko['KEYOBJECT_NAME'].instance_value())
        if type(ko_def) is dict:
            print(ko_def['COMPONENT_KEYOBJECTS'].instance_name())
            print(ko_def['COMPONENT_KEYOBJECTS'].instance_value())
        """
        # example
        # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'KM',KEYOBJECT_TYPE:'MULTIPLE',COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'KN'}}}}
        if ko['KEYOBJECT_TYPE'.lower()] == 'MULTIPLE':
            ko_name = ko['KEYOBJECT_NAME'.lower()]
            #if ko_names is not list:
            #    raise KeySqlException("__________MULTIPLE create ________________")
            #ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(ko_def['COMPONENT_KEYOBJECTS'].instance_value())
            #ko_class = ko_classes[ko_name]
            #print(ko['COMPONENT_KEYOBJECT_NAMES'])
            #print(ko['COMPONENT_KEYOBJECT_NAMES'].instance_value())
            child_name = ko['COMPONENT_KEYOBJECT_NAMES'.lower()]['KEYOBJECT_NAME'.lower()]
            #child_name = print(ko['COMPONENT_KEYOBJECT_NAMES'].instance_value()['KEYOBJECT_NAME'].instance_value())
            #print(child_name)
            #print(ko_classes)
            child = ko_classes[child_name]
            #ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(ko['COMPONENT_KEYOBJECT_NAMES'].instance_value()[0].instance_value())
            #ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(ko_classes[child_name])
            return KeyObjectMultiple(ko_name, child)

        if ko['KEYOBJECT_TYPE'.lower()] == 'COMPOSITION':
           #print(ko, '----------------------composition')
           ko_name = ko['KEYOBJECT_NAME'.lower()]

           children_instances = ko['COMPONENT_KEYOBJECT_NAMES'.lower()]
           #print(children_instances, '------------------------------ch inst')
           #print(type(children_instances))

           if type(children_instances) is dict:   # {'KEYOBJECT_NAME': instance} not list:
               children = [ko_classes[children_instances['KEYOBJECT_NAME'.lower()]]]  # KO_A1{KO_D} - compsition from one child  [type(children_instances)]
           elif type(children_instances) is list:
                #children = [ko_classes[ko_name.instance_value()] for ko_name in ko['COMPONENT_KEYOBJECT_NAMES'].instance_value()]
                #print(children_instances, '--------------------------------------x1')
                children = [ko_name["KEYOBJECT_NAME".lower()] for ko_name in children_instances]
           else:
               print("type(children_instances):", type(children_instances), '-----------------------------not implemented2')
               print(children_instances, '-----------------------------children_instances not implemented2')
               raise Exception('NOT IMPLEMENTED yet 1. (KeySqlClassFromInstance.create_keyobject_from_dict ____')

           return KeyObjectComposition(ko_name, children)

        # to catch anything else ...
        raise Exception('_____ NOT implemented yet 2. (KeySqlClassFromInstance.create_keyobject_from_dict ____')


    @staticmethod
    def create_catalog_from_instance(instance):
        # output example:
        # {'_name': 'CATALOG_DESCRIPTION', '_value':
        # {'SCHEMA_NAME': <keysql_client.keysql_instance.KeySql_Instance object at 0x000002191C643220>,
        # 'CATALOG_NAME': <keysql_client.keysql_instance.KeySql_Instance object at 0x000002191C66D730>,
        # 'STORE_NAMES': <keysql_client.keysql_instance.KeySql_Instance object at 0x000002191C66D760>,
        # 'OWNER_NAME': <keysql_client.keysql_instance.KeySql_Instance object at 0x000002191C66D7C0>,
        # 'KEYOBJECT_DEFINITIONS': <keysql_client.keysql_instance.KeySql_Instance object at 0x000002191C66D9A0>}}

        #print(type(instance),  '____________cat instance')
        #print(instance)
        if instance.instance_name() == 'CATALOG_INFO'.lower():  # catalog
            cat = instance.instance_value()
            schema_name = cat['SCHEMA_NAME'.lower()]   # {'_name': 'SCHEMA_NAME', '_value': 'TEST3'}
            cat_name = cat['CATALOG_NAME'.lower()]     # {'_name': 'CATALOG_NAME', '_value': 'CATALOG_TWIN'}
            store_names = cat['STORE_NAMES'.lower()] # {'_name': 'STORE_NAMES', '_value': None}
            if store_names is None:
                store_names = []
            elif type(store_names) is dict:
                store_names = [store_names['STORE_NAME'.lower()]]
            else:
                store_names = [rec['STORE_NAME'.lower()] for rec in store_names]
            owner_name = cat['OWNER_NAME'.lower()]  #  {'_name': 'OWNER_NAME', '_value': 'test3'}
            ko_def = cat['KEYOBJECT_DEFINITIONS'.lower()] # {'_name': 'KEYOBJECT_DEFINITIONS', '_value': {'KEYOBJECT_DEFINITION': <keysql_client.keysql_instance.KeySql_Instance object at 0x000001A4A00AEE50>}}
            if ko_def is None:
                ko_def = []
            if type(ko_def) is not list:
                ko_def = [ko_def]

            keyobjects = []
            # keyobject are in correct order ~'chronological'. To keep dependency, get them by name
            ko_classes_created = OrderedDict()
            #print(ko_def, '----------------------------------ko_def create cat')
            for ko in ko_def:
                #print(ko, '----------------------------------ko')
                ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(ko, ko_classes=ko_classes_created)
                keyobjects.append(deepcopy(ko_class))
                ko_classes_created[ko_class.get_name()] = deepcopy(ko_class)
                #print(ko_classes_created, '----------------------------------------- ko_classes_created')

            return KeySqlCatalog(cat_name, schema_name=schema_name, owner_name=owner_name, store_names=store_names, keyobject_definitions=keyobjects)

    # to compare catalog and store in keysql format
    # catalog
    # {CATALOG_DESCRIPTION:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'XYZ',STORE_NAMES:NULL,
    # OWNER_NAME:'test3',
    # KEYOBJECT_DEFINITIONS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'NN',KEYOBJECT_TYPE:'NUMBER',COMPONENT_KEYOBJECT_NAMES:NULL},
    # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'CC',KEYOBJECT_TYPE:'CHAR',COMPONENT_KEYOBJECT_NAMES:NULL},
    # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'CMPND',KEYOBJECT_TYPE:'COMPOSITION',COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'NN',KEYOBJECT_NAME:'CC'}}}}}

    # store (keysql format
    # {STORE_DESCRIPTION:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'PEOPLE',STORE_NAME:'PEOPLE_STORE',
    # OWNER_NAME:'test3',
    # HOSTS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'PERSON',KEYOBJECT_TYPE:'COMPOSITION',
    # COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'EMAILS',
    # KEYOBJECT_NAME:'FIRST_NAME',KEYOBJECT_NAME:'LAST_NAME',
    # KEYOBJECT_NAME:'YOB',KEYOBJECT_NAME:'ADDRESS'}}},
    # CONSTRAINTS:NULL}}

    # {STORE_DESCRIPTION:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'PEOPLE',STORE_NAME:'PEOPLE_STORE',
    # OWNER_NAME:'test3',
    # HOSTS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'FIRST_NAME',KEYOBJECT_TYPE:'CHAR',COMPONENT_KEYOBJECT_NAMES:NULL},
    # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'PERSON',KEYOBJECT_TYPE:'COMPOSITION',
    # COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'EMAILS',KEYOBJECT_NAME:'FIRST_NAME',KEYOBJECT_NAME:'LAST_NAME',
    # KEYOBJECT_NAME:'YOB',KEYOBJECT_NAME:'ADDRESS'}},
    # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'YOB',KEYOBJECT_TYPE:'NUMBER',COMPONENT_KEYOBJECT_NAMES:NULL}},CONSTRAINTS:NULL}}


    # instance: (after mapping store -> instance
    # {'SCHEMA_NAME': <keysql_client.keysql_instance.KeySql_Instance object at 0x0000021C8BB27190>,
    # 'CATALOG_NAME': <keysql_client.keysql_instance.KeySql_Instance object at 0x0000021C8BD98DC0>,
    # 'STORE_NAME': <keysql_client.keysql_instance.KeySql_Instance object at 0x0000021C8BDA1B80>,
    # 'OWNER_NAME': <keysql_client.keysql_instance.KeySql_Instance object at 0x0000021C8C115F40>,
    # 'HOSTS': <keysql_client.keysql_instance.KeySql_Instance object at 0x0000021C8C12B370>,
    # 'CONSTRAINTS': <keysql_client.keysql_instance.KeySql_Instance object at 0x0000021C8C12B3D0>}

    @staticmethod
    def create_store_from_instance(instance):
        #print(type(instance),  '____________')
        if instance.instance_name() == 'STORE_INFO'.lower():  # catalog
            store = instance.instance_value()
            print(store)
            print('----------------------------------------xxxx start')

            schema_name = store['SCHEMA_NAME'.lower()]   # {'_name': 'SCHEMA_NAME', '_value': 'TEST3'}
            cat_name = store['CATALOG_NAME'.lower()]     # {'_name': 'CATALOG_NAME', '_value': 'CATALOG_TWIN'}
            store_name = store['STORE_NAME'.lower()]     # {'_name': 'STORE_NAME', '_value': None}

            owner_name = store['OWNER_NAME'.lower()]  #  {'_name': 'OWNER_NAME', '_value': 'test3'}
            ko_hosts = store['HOSTS'.lower()]   # {'_name': 'KEYOBJECT_DEFINITIONS', '_value': {'KEYOBJECT_DEFINITION': <keysql_client.keysql_instance.KeySql_Instance object at 0x000001A4A00AEE50>}}
            if ko_hosts is None:
                ko_hosts = []
            if type(ko_hosts) is not list:
                ko_hosts = [ko_hosts]

            #print(ko_hosts, '------------------------------')
            ko_host_names = []
            if len(ko_hosts) > 0:
                if type(ko_hosts[0]) is dict:
                    ko_host_names = [ko['KEYOBJECT_DEFINITION'.lower()]['KEYOBJECT_NAME'.lower()] for ko in ko_hosts]
                    #print(ko_host_names, '------------------------------- ko_host_names from dict')
                elif type(ko_hosts[0]) is KeySql_Instance:
                    ko_host_names = [ko['KEYOBJECT_NAME'.lower()] for ko in ko_hosts]
                    #print(ko_host_names, '------------------------------- ko_host_names from instance')
                #ko_host_names = [ko.instance_value()['KEYOBJECT_DEFINITION'].instance['KEYOBJECT_NAME'].instance_value() for ko in ko_hosts]

                #ko_host_names = [ko.instance_value()['KEYOBJECT_NAME'].instance_value() for ko in ko_hosts]
                #print(ko_host_names)
                #print(ko_hosts, '------------------------------x')
                #ko_host_names = [ko.instance_value()['KEYOBJECT_NAME'].instance_value() for ko in ko_hosts]
            #print(ko_host_names, '------------------------------------------------ko_host_names -> create_store_from_instance')

            # instances until next iteration
            #ko_hosts = [ko['KEYOBJECT_DEFINITION'].instance_value() for ko in ko_hosts]
            #print(ko_hosts, '-----------------------------------------------------ko_hosts -> create_store_from_instance')
            # TODO:
            #  to get host keyobject, need to get all components from catalog, maybe todo it later
            # now: return anly keyobject names
            '''
            keyobjects = []
            # keyobject are in correct order ~'chronological'. To keep dependency, get them by name
            ko_classes_created = OrderedDict()
            print(ko_hosts, '----------------------------------ko_def create STORE')
            for ko in ko_hosts:
                print(ko, '----------------------------------ko create store from instance')
                # for classes:
                #ko_class = KeySqlClassFromInstance.create_keyobject_from_instance(ko, ko_classes=ko_classes_created)
                #keyobjects.append(deepcopy(ko_class))
                #ko_classes_created[ko_class.get_name()] = deepcopy(ko_class)
                print(ko_classes_created, '----------------------------------------- ko_classes_created')
            '''

            # COMPONENT_KEYOBJECT_NAMES
            #print(store, '----------------------------')
            #print(ko_hosts)
            comp_names = []
            if len(ko_hosts) > 0 and type(ko_hosts[0]) is dict:
                ko_comp = [ko['KEYOBJECT_DEFINITION'.lower()]['COMPONENT_KEYOBJECT_NAMES'.lower()] for ko in ko_hosts]
                # is still instances
                comp_names = [[kk for kk in ko] if type(ko) is list else [] for ko in ko_comp]
                #print(comp_names, '------------------------------- comp_names from dict')
            elif len(ko_hosts) > 0 and type(ko_hosts[0]) is KeySql_Instance:
                ko_comp = [ko['COMPONENT_KEYOBJECT_NAMES'.lower()] for ko in ko_hosts]
                comp_names = [[kk for kk in ko] if type(ko) is list else [] for ko in ko_comp]
                #print(comp_names, '------------------------------- comp_names from instance')

            # CONSTRAINTS
            comp_constrains = store['CONSTRAINTS'.lower()]
            if comp_constrains is None:
                comp_constrains = []
            if type(comp_constrains) is not list:
                comp_constrains = [comp_constrains]

            # KeySqlStore: __init__(self, name: str, catalog_name: str, schema_name="", owner_name="", host_names=[], hosts=[], constraints=[]):
            return KeySqlStore( store_name, cat_name, schema_name=schema_name, owner_name=owner_name, host_names=ko_host_names, hosts=ko_hosts, component_names=comp_names, constraints=comp_constrains)


if __name__ == '__main__':

    # store example
    store_keysql = "{STORE_INFO:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'PEOPLE',STORE_NAME:'PEOPLE_STORE',OWNER_NAME:'test3',HOSTS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'PERSON',KEYOBJECT_TYPE:'COMPOSITION',COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'EMAILS',KEYOBJECT_NAME:'FIRST_NAME',KEYOBJECT_NAME:'LAST_NAME',KEYOBJECT_NAME:'YOB',KEYOBJECT_NAME:'ADDRESS'}}},CONSTRAINTS:NULL}}"
    instance = KeySqlParser.parse_select_reply(store_keysql)[0]

    store = KeySqlClassFromInstance.create_store_from_instance(instance)

    keysql_str = "{CATALOG_INFO:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'XYZ',STORE_NAMES:NULL,OWNER_NAME:'test3',KEYOBJECT_DEFINITIONS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'NN',KEYOBJECT_TYPE:'NUMBER',COMPONENT_KEYOBJECT_NAMES:NULL},KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'CC',KEYOBJECT_TYPE:'CHAR',COMPONENT_KEYOBJECT_NAMES:NULL},KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'CMPND',KEYOBJECT_TYPE:'COMPOSITION',COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'NN',KEYOBJECT_NAME:'CC'}}}}}"

    instance = KeySqlParser.parse_select_reply(keysql_str)[0]

    cat = KeySqlClassFromInstance.create_catalog_from_instance(instance)
