from keysql_client.keysql_connection_async import KeySqlConnectionAsync

from tornado.ioloop import IOLoop

class KeySqlClientAsync:
    def __init__(self, host: str, port: int, ioloop=IOLoop.current()):
        """
        KeySqlClient for asyncronious running
        :param host: host - string
        :param port: port - integer
        :param ioloop:  IOLoop.current() - default
        :param conn_type:  'async' - default (for native python, 'async_coroutine - for tornado coroutine
        """
        self._host = host
        self._port = port
        self._ioloop = ioloop
        self._connection = KeySqlConnectionAsync(self._host, self._port, self._ioloop)

    #def connect(self, username: str, password: str, secure_ssl=True):
    #    return KeySqlConnection(self._host,self._port,username,password,secure_ssl)

    async def connect(self, username: str, password: str, ssl=True):
        await self.connect_base(username, password, ssl=ssl)
        return self._connection

    async def connect_base(self, username: str, password: str, ssl=True):
        """
        Connect with server using username and password
        :param username:  user name, string
        :param password:  password, string
        :param ssl: default: true (secure connection), False - non-secure connection
        :return:
        """
        await self._connection.connect_user(username, password, ssl=ssl)

