#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

#from keysql_parser import *
from keysql_client.keysql_exception import KeySqlExceptionMessage, KeySqlException
import logging


#   # example
#  {'KEYOBJECT_DEFINITION': {'KEYOBJECT_NAME': 'KO1', 'KEYOBJECT_TYPE': 'NUMBER', 'COMPONENT_KEYOBJECT_NAMES': None}}
class KeySql_Instance(object):
    _name : str
    def __init__(self, name, value=None, keyobject_type="UNDEFINED", component_names=[]):    # input_list_dict_instance):
        self._name = name
        self._value = value

        #self._input_list_dict_instance = value   #  input_list_dict_instance
        #self._value = self.process_any(value)   #  input_list_dict_instance

    def keyobject_type(self):
        pass

    def instance_name(self):
        return self._name   # self.__name__

    def instance_value(self):
        return self._value

    def process_any(self, any_inst):
        if type(any_inst) == dict:
            return self.process_dict(any_inst)
        return any_inst

    def process_dict(self, any_inst):
        for key in any_inst.keys():
            any_inst[key] = self.process_any(any_inst[key])
        return any_inst

#    def process_reply(self, reply):
#         return KeySqlParser.parse_select_reply(reply)

    @staticmethod
    def instances_to_string(list_or_inst):
        if type(list_or_inst) == list:
            res = [KeySql_Instance.instances_to_string(item) for item in list_or_inst]
            #return "[" + ",".join(res) + "]"
            return "{" + ",".join(res) + "}"

        if type(list_or_inst) == KeySql_Instance:

            #res = "{" + str(list_or_inst.__name__) + ":" + KeySql_Instance.instances_to_string(list_or_inst._value) + "}"
            res = "{" + list_or_inst._name + ":" + KeySql_Instance.instances_to_string(list_or_inst._value) + "}"
            return res

        if hasattr(list_or_inst, '__dict__'):
            dict_list_or_inst = list_or_inst.__dict__
            res = "{" + list_or_inst.__class__.__name__ + ":" + KeySql_Instance.instances_to_string(dict_list_or_inst) + "}"
            return res

        if type(list_or_inst) == dict:
            res = []
            for key in list_or_inst.keys():
                val = list_or_inst[key]
                if val is None:
                    val = 'NULL' # str(val)  # ????
                elif type(val) == str:
                    # create double quote
                    val_replace = str(val).replace("\'", "\'\'")
                    #val = "'" + str(val) + "'"
                    val = "'" + val_replace + "'"
                elif type(val) in [int, float]:
                    val = str(val)

                elif type(val) == list:
                    #val = KeySql_Instance.instances_to_string(val)
                    #return "{" + key + ":" + ",".join(val) + "}"
                    #return key + ":" + "{" + val + "}"
                    val = [KeySql_Instance.instances_to_string(item) for item in val]
                    val = [item[1:-1] for item in val]  # remove { } for each item in list   ????

                    val = "{" + ",".join(val) + "}"

                elif type(val) == KeySql_Instance:   #  in  [KeySql_Instance, dict]:
                    val = KeySql_Instance.instances_to_string(val)

                elif type(val) == dict:
                    val = KeySql_Instance.instances_to_string(val)
                else:
                    val = KeySql_Instance.instances_to_string(val.__dict__)


                res.append(key + ":" + val)
            #return "{" + key + ":" + ",".join(res) + "}"
            return "{" + ",".join(res) + "}"
            #return ",".join(res)

        else:
            logging.error('Error:', type(list_or_inst), list_or_inst )
            raise KeySqlException(KeySqlExceptionMessage.DATA_TYPE_IS_NOT_PROCESSED)

    @staticmethod
    def equal_two_instances_unordered(instance1, instance2):
        if instance1 is None and instance2 is None:
            return True
        if instance1 is None or instance2 is None:
            return False

        if type(instance1) != type(instance2):
            return False

        if type(instance1) in [int, float, str]:
            return instance1 == instance2

        if type(instance1) is dict:
            if len(instance1.keys()) != len(instance2.keys()):
                return False
            keys2 = instance2.keys()
            for kk in instance1.keys():
                if kk not in keys2:
                    return False
                if not KeySql_Instance.equals_unorded(instance1[kk], instance2[kk]):
                    return False
            return True

        return Exception("KeySql_Instance.compare_one_vs_one_instances: not implemented yet")


    @staticmethod
    def equals_unorded(instance1, instance2):
        if type(instance1) != type(instance2):
            return False

        if type(instance1) in [int, float, str]:
            return instance1 == instance2

        if type(instance1) == list:
            if len(instance1) != len(instance2):
                return False

            used = [False]*len(instance1)
            for i in range(len(instance1)):
                instance_exists = False
                for j in range(len(instance2)):
                    if not used[j] and KeySql_Instance.equal_two_instances_unordered(instance1[i], instance2[j]):
                        used[j] = True
                        instance_exists = True
                        break
                if not instance_exists:
                    return False

            return True
        return Exception("KeySql_Instance.compare_two_instances: not implemented yet")


    @staticmethod
    def equals_unorded(instance1, instance2):
        if type(instance1) != type(instance2):
            return False

        if type(instance1) in [int, float, str]:
            return instance1 == instance2

        if type(instance1) == list:
            if len(instance1) != len(instance2):
                return False

            used = [False]*len(instance1)
            for i in range(len(instance1)):
                instance_exists = False
                for j in range(len(instance2)):
                    if not used[j] and KeySql_Instance.equal_two_instances_unordered(instance1[i], instance2[j]):
                        used[j] = True
                        instance_exists = True
                        break
                if not instance_exists:
                    return False

            return True

        return Exception("KeySql_Instance.compare_two_instances: not implemented yet")


    @staticmethod
    def equal_two_instances(instance1, instance2):
        if instance1 is None and instance2 is None:
            return True
        if instance1 is None or instance2 is None:
            return False

        if type(instance1) != type(instance2):
            return False

        if type(instance1) in [int, float, str]:
            return instance1 == instance2

        if type(instance1) is dict:
            if len(instance1.keys()) != len(instance2.keys()):
                return False
            keys2 = instance2.keys()
            for kk in instance1.keys():
                if kk not in keys2:
                    return False
                if not KeySql_Instance.equals(instance1[kk], instance2[kk]):
                    return False
            return True
        return Exception("KeySql_Instance.compare_one_vs_one_instances: not implemented yet")


    @staticmethod
    def equals(instance1, instance2):
        if type(instance1) != type(instance2):
            return False

        if type(instance1) in [int, float, str]:
            return instance1 == instance2

        if type(instance1) == list:
            if len(instance1) != len(instance2):
                return False

            for i in range(len(instance1)):
               if not KeySql_Instance.equal_two_instances(instance1[i], instance2[i]):
                   return False
            return True

        return Exception("KeySql_Instance.compare_two_instances: not implemented yet")


    def serialize_old(self):
        if type(self._value) == list:   # _input_list_dict_instance
            res = [self.process_dict(dict_inst) for dict_inst in self._value]            # _input_list_dict_instance
            res = ','.join(res)
        if type(self._value) == dict:    # _input_list_dict_instance
            key = self._value.keys()[0]                                                  # _input_list_dict_instance__
            res = self.process_dict(self._value[key])                                    # self._input_list_dict_instance
            res = '{' + str(key) + ':' + res + '}'  # str(key) + ':' + '{' + res + '}'

        elif type((self._value)) == list:
            res = [self.process_dict(item) for item in self._value]
            res = '{' + str(self._name) + ':[' + ','.join(res) + ']}'
        return res



# 'Null'
if __name__ == "__main__":

  seq = "INSERT INTO Billiard_Store INSTANCES\
    {BallPack: {Brand: 'Carambo', Name: 'Standard_Carom' , UPC: 123456789012,\
         BallSet: { Style:'carom', Diam_mm: 61.5, Qty: 3,\
                       Balls: { Ball:{ Num: Null, Color:'red'},\
                                Ball:{ Num: Null, Color:'white'},\
                                Ball:{ Num: Null, Color:'white_spot'}\
                              }}} }"

  big_example =[
    {'BallPack': {'Brand': 'Carambo', 'Name': 'Standard_Carom' , 'UPC': 123456789012,
     'BallSet': {'Style':'carom', 'Diam_mm': 61.5, 'Qty': 3,
                 'Balls':  [ {'Ball':{ 'Num': 'Null', 'Color':'red'}},
                             {'Ball':{ 'Num': 'Null', 'Color':'white'}},
                             {'Ball':{ 'Num': 'Null', 'Color':'white_spot'}}
	                        ]
                }
        }
    },
{'BallPack': {'Brand': 'Gorda', 'Name': 'Deluxe' , 'UPC': 109234567812,
     'BallSet': { 'Style':'pool', 'Diam_mm': 57.15, 'Qty': 15,
              'Balls': [ { 'Ball':{ 'Num': 1,   'Color':'yellow'}},
                         {  'Ball':{ 'Num': 2,   'Color':'blue'}},
                         {   'Ball':{ 'Num': 3,   'Color':'red'}},
                         {   'Ball':{ 'Num': 4,   'Color':'pink'}},
                         {   'Ball':{ 'Num': 5,   'Color':'orange'}},
                         {   'Ball':{ 'Num': 6,   'Color':'green'}},
                         {   'Ball':{ 'Num': 7,   'Color':'tan'}},
                         {   'Ball':{ 'Num': 8,   'Color':'black'}},
                         {   'Ball':{ 'Num': 9,   'Color':'yellow_white'}},
                         {   'Ball':{ 'Num': 10,  'Color':'blue_white'}},
                         {   'Ball':{ 'Num': 11,  'Color':'red_white'}},
                         {   'Ball':{ 'Num': 12,  'Color':'pink_white'}},
                         {   'Ball':{ 'Num': 13,  'Color':'orange_white'}},
                         {   'Ball':{ 'Num': 14,  'Color':'green_white'}},
                         {   'Ball':{ 'Num': 15,  'Color':'tan_white'}}
                       ]
			  }
	     }
},
{'BallPack': {'Brand': 'Pobeda', 'Name': 'Atos' , 'UPC': 345120967812,
     'BallSet': { 'Style':'pyramid', 'Diam_mm': 68, 'Qty': 16,
                   'Balls': [ {'Ball':{ 'Num': 1,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 2,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 3,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 4,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 5,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 6,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 7,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 8,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 9,    'Color':'ivory'}},
                              {'Ball':{ 'Num': 10,   'Color':'ivory'}},
                              {'Ball':{ 'Num': 11,   'Color':'ivory'}},
                              {'Ball':{ 'Num': 12,   'Color':'ivory'}},
                              {'Ball':{ 'Num': 13,   'Color':'ivory'}},
                              {'Ball':{ 'Num': 14,   'Color':'ivory'}},
                              {'Ball':{ 'Num': 15,   'Color':'ivory'}},
                              {'Ball':{ 'Num': 'Null', 'Color':'yellow'}}
                  ]
		     }
	}
},
{'BallPack': {'Brand': 'Safo', 'Name': 'Super' , 'UPC': 678345120912,
     'BallSet': { 'Style': 'snooker', 'Diam_mm': 52.5, 'Qty': 7,
                   'Balls': [{'Ball':{ 'Num': 'Null',  'Color':'red'}},
                             {'Ball':{ 'Num': 'Null',  'Color':'yellow'}},
                             {'Ball':{ 'Num': 'Null',  'Color':'green'}},
                             {'Ball':{ 'Num': 'Null',  'Color':'brown'}},
                             {'Ball':{ 'Num': 'Null',  'Color':'blue'}},
                             {'Ball':{ 'Num': 'Null',  'Color':'pink'}},
                             {'Ball':{ 'Num': 'Null',  'Color':'black'}}
                          ]
			}
	}
 }
]

  print(type(big_example))
  print(type(big_example[0]))

  print(big_example)

  print('abc')
  print(str("'abc'"))

  ki = KeySql_Instance('test', big_example)
  #res = ki.serialize()
  #print(res)
  print(ki)


