#/**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/

import logging
#from keysql_client.keysql_connector import *
from keysql_client.keysql_parser import KeySqlParser
from keysql_client.keysql_instance import KeySql_Instance
from keysql_client.key_object import KeyObject
from keysql_client.key_object import KeyObjectChar, KeyObjectNumber, KeyObjectDate, KeyObjectMultiple
from keysql_client.keysql_catalog import KeySqlCatalog

#TODO: delete -> remove later
from keysql_client.keysql_mapper import KeySqlMapper

class KeySqlMapperJson(object):

    def __init__(self):
        pass

    @staticmethod
    def class_to_instance(class_inst) -> KeySql_Instance:
        #TODO
        pass

    @staticmethod
    def create_keyobject_from_dict(ko_def):
        #TODO !!!!!!!!!!
        # example
        # {'KEYOBJECT_DEFINITION': {'KEYOBJECT_NAME': 'KO1', 'KEYOBJECT_TYPE': 'NUMBER', 'COMPONENT_KEYOBJECT_NAMES': None}}
        if ko_def['KEYOBJECT_DEFINITION']['KEYOBJECT_TYPE'] == 'NUMBER':
            return KeyObjectNumber(ko_def['KEYOBJECT_DEFINITION']['KEYOBJECT_NAME'])

        # {'KEYOBJECT_DEFINITION': {'KEYOBJECT_NAME': 'KO2', 'KEYOBJECT_TYPE': 'CHAR', 'COMPONENT_KEYOBJECT_NAMES': None}}
        if ko_def['KEYOBJECT_DEFINITION']['KEYOBJECT_TYPE'] == 'CHAR':
            return KeyObjectChar(ko_def['KEYOBJECT_DEFINITION']['KEYOBJECT_NAME'])

        if ko_def['KEYOBJECT_DEFINITION']['KEYOBJECT_TYPE'] == 'DATE':
            return KeyObjectDate(ko_def['KEYOBJECT_DEFINITION']['KEYOBJECT_NAME'])

        if ko_def['KEYOBJECT_DEFINITION']['KEYOBJECT_TYPE'] == 'MULTIPLE':
            return KeyObjectMultiple(ko_def['KEYOBJECT_DEFINITION']['KEYOBJECT_NAME'], ko_def['KEYOBJECT_DEFINITION']['COMPONENT_KEYOBJECT_NAMES'])

        # to catch anything else
        raise Exception('_____ NOT implemented  create_keyobject_from_dict ____')

    # dict looks like
    # {"CATALOG_DESCRIPTION":{"SCHEMA_NAME":"TEST3","CATALOG_NAME":"CATALOG02","STORE_NAMES":null,
    # "OWNER_NAME":"test3","KEYOBJECT_DEFINITIONS":null}}

    @staticmethod
    def create_catalog_from_dict(cat_dict):
        keyobject_definitions = cat_dict["CATALOG_DESCRIPTION"]["KEYOBJECT_DEFINITIONS"]
        if keyobject_definitions is None:
            keyobjects = []
        else:
            # transform keyobjects
            keyobject_definitions = []
            for ko_def in cat_dict["CATALOG_DESCRIPTION"]["KEYOBJECT_DEFINITIONS"]:
                ko = KeySqlMapperJson.create_keyobject_from_dict(ko_def)
                keyobject_definitions.append(ko)

        store_names = cat_dict["CATALOG_DESCRIPTION"]["STORE_NAMES"]
        if store_names is None:
            store_names = []

        schema_name = cat_dict["CATALOG_DESCRIPTION"]["SCHEMA_NAME"]
        catalog_name = cat_dict["CATALOG_DESCRIPTION"]["CATALOG_NAME"]
        owner_name = cat_dict["CATALOG_DESCRIPTION"]["OWNER_NAME"]

        return KeySqlCatalog(catalog_name, schema_name, owner_name, store_names, keyobject_definitions)


    #--------------------------------------------------------------------------------------------------------------


    @staticmethod
    def keysql_inst_to_dictionary_only(class_inst):

        if class_inst.__class__ == list:
            return list(map(KeySqlMapper.keysql_inst_to_dictionary_only, class_inst))
        if type(class_inst) == KeySql_Instance:
            res = class_inst.__dict__
            if res['__name__'] == 'RESULT':
                res = KeySqlMapper.keysql_inst_to_dictionary_only(res['_value'])
                return res
            #TODO
            # to uncomment if remove #BALL (example)
            if res['__name__'][0] == '#':   # .startwith('#'):
                return KeySqlMapper.keysql_inst_to_dictionary_only(res['_value'])
                #return list(map(KeySqlMapper.keysql_inst_to_dictionary_only, res['_value']))

            return {class_inst.__dict__['__name__'] : KeySqlMapper.keysql_inst_to_dictionary_only(class_inst.__dict__['_value'])}

        if class_inst.__class__ == dict:
            if 'RESULT' in class_inst.keys():
                return class_inst['_value']
            if len(class_inst.keys()) == 1 and list(class_inst.keys())[0].startswith('#'):
                return KeySqlMapper.keysql_inst_to_dictionary_only(class_inst.values()[0])

            res = class_inst
            for key in res.keys():
                if key.startswith('#'):
                    return KeySqlMapper.keysql_inst_to_dictionary_only(res[key])
                if res[key].__class__ == KeySql_Instance:  #  dict:
                    res[key] = KeySqlMapper.keysql_inst_to_dictionary_only(res[key])

            return res
            #return list(map(KeySqlMapper.keysql_inst_to_dictionary_only, res['_value']))

        return class_inst


    @staticmethod
    def keysql_inst_to_dictionary_only_backup(class_inst):

        if class_inst.__class__ == list:
            return list(map(KeySqlMapper.keysql_inst_to_dictionary_only, class_inst))
        if type(class_inst) == KeySql_Instance:
            res = class_inst.__dict__
            if res['__name__'] == 'RESULT':
                res = tuple(KeySqlMapper.keysql_inst_to_dictionary_only(res['_value']))
                return res
            #TODO
            # to uncomment if remove #BALL (example)
            if res['__name__'][0] == '#':   # .startwith('#'):
                return KeySqlMapper.keysql_inst_to_dictionary_only(res['_value'])
                #return tuple(map(KeySqlMapper.keysql_inst_to_dictionary_only, res['_value']))

            return {class_inst.__dict__['__name__'] : KeySqlMapper.keysql_inst_to_dictionary_only(class_inst.__dict__['_value'])}

        if class_inst.__class__ == dict:
            #if 'RESULT' in class_inst.keys():
            #    return class_inst['_value']
            if len(class_inst.keys()) == 1 and class_inst.keys()[0].startswith('#'):
                return KeySqlMapper.keysql_inst_to_dictionary_only(class_inst.values()[0])

            res = class_inst
            for key in res.keys():
                if key.startswith('#'):
                    return KeySqlMapper.keysql_inst_to_dictionary_only(res[key])
                if res[key].__class__ == KeySql_Instance:  #  dict:
                    res[key] = KeySqlMapper.keysql_inst_to_dictionary_only(res[key])

        return class_inst
    '''
    Robot2 = type("Robot2", 
                  (), 
                  {"counter":0, 
                   "__init__": Rob_init,
                   "sayHello": lambda self: "Hi, I am " + self.name})
    '''


    #def get_default_class_from_dict(dict_created, base_classes = []):
    #    keys = dict_created.keys()
    #    # the only key
    #    key = list(keys)[0]
    #    return type(key, (), dict_created[key])

    @staticmethod
    def get_default_class_from_dict(dict_created, name, base_classes = []):

        #keys = list(dict_created.keys())
        # the only key
        #key = keys[0]
        return type(name, (), dict_created)

    @staticmethod
    def get_default_class_from_instance(instance,  base_classes = []):
        #keys = dict_created.keys()
        # the only key
        #key = instance['__name__']
        key = instance._name
        #res = type(key, (object,), instance['_value'])

        if type(instance._value) in [int, float, str] or instance._value is None:
            #simple_obj = KeySqlMapper.get_default_class_from_instance(class_inst)
            return instance._name

        if len(base_classes) == 0:
            res = type(key, (object,), instance._value)   #   res = type(key, (object,), instance)
        else:
            res = type(key, tuple(base_classes), instance._value)

        return res  #  type(key, (), instance['_value'])


    @staticmethod
    def keysql_inst_to_default_class_only(class_inst):

        if class_inst.__class__ == list:
            return list(map(KeySqlMapper.keysql_inst_to_default_class_only, class_inst))

        if class_inst._value.__class__ == list:
            return list(map(KeySqlMapper.keysql_inst_to_default_class_only, class_inst._value))

        if not class_inst._value or class_inst._value.__class__ in (int,float,str):
            return class_inst._value

        # logging.debug(class_inst.__name__ + ": type" + str(class_inst._value.__class__))
        assert class_inst._value.__class__ == dict

        r = {}
        for key, value in class_inst._value.items():
            r[key] = KeySqlMapper.keysql_inst_to_default_class_only(value)

        return KeySqlMapper.get_default_class_from_dict(r,class_inst._name)

    @staticmethod
    def keysql_inst_to_dict_impl(class_inst,top):

        if class_inst.__class__ == list:
            r = {}
            for x in class_inst:
                r[class_inst._name] = KeySqlMapper.keysql_inst_to_dict_impl(x, False)
            return r

        if class_inst._value.__class__ == list:
            if class_inst._name == 'RESULT':
                r = {}
                for x in class_inst._value:
                    r[x._name] = KeySqlMapper.keysql_inst_to_dict_impl(x, False)
                return r
            else:
                r = []
                for x in class_inst._value:
                    r.append(KeySqlMapper.keysql_inst_to_dict_impl(x,False))
                return r

        if not class_inst._value or class_inst._value.__class__ in (int, float, str):
            return class_inst._value

        assert class_inst._value.__class__ == dict
        r = {}
        for key, value in class_inst._value.items():
            r[key] = KeySqlMapper.keysql_inst_to_dict_impl(value,False)
        if top:
            return {class_inst._name : r}
        return r

    @staticmethod
    def keysql_inst_to_dict(class_inst):
        return KeySqlMapper.keysql_inst_to_dict_impl(class_inst,True)

    @staticmethod
    def keysql_inst_to_default_class(class_inst):
        return KeySqlMapper.keysql_inst_to_default_class_only(class_inst)

    #@classmethod
    @staticmethod
    def keysql_inst_to_class_inst(class_inst, class_types:list=[]):
        logging.debug("in keysql_inst_to_class_inst: ", str(class_types))

        if type(class_inst) == list:
            res = []
            for item in class_inst:
                res.append(KeySqlMapper.keysql_inst_to_class_inst(item, class_types))
            return res
        #if hasattr(class_inst, '__name__') and class_inst['__name__'][0] == '#':
        #    res = KeySqlMapper.class_inst_to_dict(class_inst['__name__'])

        elif hasattr(class_inst, '__dict__'):
            res = class_inst.__dict__
            #print('keys:', res.keys)
            #print(class_inst.__repr__())
            for key in res.keys():
                #print('key:', key)
                #print(class_inst.__class__.__name__, '-------\n')
                if hasattr(res[key], '__dict__'):
                    class_inst[key] = KeySqlMapper.keysql_inst_to_class_inst(class_inst[key], class_types)

                elif type(res[key]) == list:
                    processed = res[key]
                    #for i in range(len(processed)):
                    #    processed[i] = KeySqlMapper.class_inst_to_dict(processed[i])
                    res = KeySqlMapper.keysql_inst_to_class_inst(res[key])
                    if key[0] != '#':
                        res = {key:res}
                    #res[key] = processed

                     #return res # #return class_inst
        else:
            res = [class_inst]
            #name = class_inst.__class__.__name__
            #return [{name: res}]
        #return [{name: res}]
        # result of different select and join
        if hasattr(class_inst, '__name__'):
            if res['__name__'] == 'RESULT':
                res = res['_value']
            # returns list of instansces
            elif res['__name__'] is not None and res['__name__'][0] == '#':
                res = res['_value']
        return res


def dump(obj, level=0):

    for attr in dir(obj):
      val = getattr(obj, attr)
      print('val:', val)
      #print('val:', val.)
      #if issubclass(val, (int, float, str, socket, list, dict, set)):
      if issubclass(val, socket.socket.__class__  (type(socket))):
              print(level*' ', val)
      else:
           dump(val, level=level+1)

# if hasattr(a, 'property'):
#     doStuff(a.property)
# else:
#     otherStuff()

def dump_obj_prev(obj, level=0):
    print(obj)

    if hasattr(obj, '__dict__'):
        print('all_items: ', obj.__dict__.items())

        for key, value in obj.__dict__.items():
            #print('___key->val:', key, value, type(value))

            #if isinstance(value, (int, float, str, list, dict, set, socket)):
            if isinstance(value, (int, float, str, list, dict, set)):
                print(" " * level + "%s -> %s" % (key, value))
            else:
                print(" " * level + "%s -> %s" % (key, value))
                #print('vars:', vars(value))
                dump_obj(value, level + 2)
    else:
        print('no atts:', obj)


#keyobject
def dump_obj(obj, level=0):
    print(obj)

    if type(obj) == KeyObject:  # hasattr(obj, '__dict__'):
        print('all_items: ', obj.__dict__.items())

        for key, value in obj.__dict__.items():
            print('___key->val:', key, value, type(value))

            #if isinstance(value, (int, float, str, list, dict, set, socket)):
            if isinstance(value, (int, float, str, list, dict, set)):
                print(" " * level + "%s -> %s" % (key, value))
            else:
                print(" " * level + "%s -> %s" % (key, value))
                #print('vars:', vars(value))
                dump_obj(value, level + 2)
    else:
        print('no atts:', obj)


#@annotation keyobject_name='Ballyy'

'''
example
def decorator(arg1, arg2):
    def real_decorator(function):
        def wrapper(*args, **kwargs):
            print "Congratulations.  You decorated a function that does something with %s and %s" % (arg1, arg2)
            function(*args, **kwargs)
        return wrapper
    return real_decorator
'''

def decorator_class_name(arg):
    def real_decorator(function):
        def wrapper(*args, **kwargs):
            #print("Congratulations.  You decorated a function that does something with %s" % (arg)
            function(*args, **kwargs)
        return wrapper
    return real_decorator

#@decorator_class_name("Ball1")
class Ball:
    def __init__(self, num, color, alias='None'):
        self.color = color
        self.num = num
        self.alias = property(lambda self: alias)

#class Balls:
#    def __init__(self, balls):
#        self.balls = balls

class BallSet:
    def __init__(self, style, diam_mm, qty, balls):
        self.style = style
        self.diam_mm = diam_mm
        self.qty = qty
        self.balls = balls


# only for dictionaries
class AllMyFields:
    def __init__(self, dictionary):
        for k, v in dictionary.items():
            setattr(self, k, v)


if __name__ == "__main__":

    str1 = "{LAST_NAME:\'Johns\'}"
    test1 = KeySqlParser.parse_select_reply(str1)
    print(test1)
    print(type(test1))
    print(test1[0])
    assert type(test1[0]) == KeySql_Instance # # assert test1 == "['Johns']"
    res1 = KeySqlMapper.keysql_inst_to_default_class_only(test1)
    print(res1)
    print(dir(res1))

    str1 = "{EMAIL:{EMAIL_ADDRESS:\'shopping@johns.com\',EMAIL_TYPE:NULL}}"
    test1 = KeySqlParser.parse_select_reply(str1)
    print(test1)
    res1 = KeySqlMapper.keysql_inst_to_default_class_only(test1)
    print(res1)
    print(dir(res1))


    str1 = "{EMAILS:{EMAIL:{EMAIL_ADDRESS:\'shopping@johns.com\',EMAIL_TYPE:NULL}},FIRST_NAME:NULL,LAST_NAME:\'Johns\',YOB:1999}"
    str2 = "{PERSON:{EMAILS:{EMAIL:{EMAIL_ADDRESS:\'shopping@johns.com\',EMAIL_TYPE:NULL}},FIRST_NAME:NULL,LAST_NAME:\'Johns\',YOB:1999}}"

    """
    small_example = "{BallPack: {Brand: 'Carambo', Name: 'Standard_Carom' , UPC: 123456789012, BallSet: { Style:'carom', Diam_mm: 61.5, Qty: 3, Balls: { Ball:{ Num: Null, Color:'red'},Ball:{ Num: Null, Color:'white'},}}}}"

    print(small_example)

    small_example = small_example.replace(' ', '')
    print(small_example)

    res = small_example.split('}')
    res = [item + '}' for item in res]
    for item in res:
        print(item)


    print({'a':'b'})
    print({'a':{'b':[{'c':'d'}, {'e':'f'}]}})

    o =  AllMyFields({'a': 1, 'b': 2})
    print(o.a, o.b)
    print(o)


    ball_1 = Ball(None, 'red')
    ball_2 = Ball(None, 'white')
    ball_3 = Ball(None, 'white_spot')

    #all_1.alias('ball_1')

    print('\nball:')
    print(ball_1.__dict__)
    print(vars(ball_1))
    for elem in vars(ball_1):
        print(elem, type(elem))

    print(ball_2)
    print(ball_3)
    balls = [ball_1, ball_2, ball_3]

    #print('\nballs:')
    #print(vars(balls))

    ballset = BallSet('carom', 61.5, 3, balls)

    print('\nballset:')
    print(dir(ballset))
    print(vars(ballset))

    print('\nballset.__dict__:')
    print(ballset.__dict__)
    print(ballset.__dict__['balls'][0])    # [0].__dict__)
    print(ballset.__dict__['balls'][0])    # [0].__dict__)
    print(type(ballset.__dict__['balls'][0]))

    print(type(ballset.__dict__['balls'][0].__dict__))
    print(ballset.__dict__['balls'][0].__dict__)

    ballset_to_dict = KeySqlMapper.class_inst_to_dict(ballset)

    print(ballset_to_dict)


    print('-'*40)

    for elem in vars(ballset):
        print(elem, type(elem))

    # if hasattr(a, 'property'):
    #     a.property

    dict_a = {'BallPack': {'Brand': 'Carambo', 'Name': 'Standard_Carom' , 'UPC': 123456789012,
     'BallSet': { 'Style':'carom', 'Diam_mm': 61.5, 'Qty': 3,
                   'Balls':  [{'Ball':{ 'Num': 'Null', 'Color':'red'}},
                               {'Ball':{ 'Num': 'Null', 'Color':'white'}},
                                {'Ball':{ 'Num': 'Null', 'Color':'white_spot'}}]
     	   	      }
           }
}
    map_class = AllMyFields(dict_a)

    #from bunch import bunchify

    print(vars(map_class))

    print(map_class)

    print('-'*40)
    d = {'a': 1, 'b': {'c': 2}, 'd': ["hi", {'foo': "bar"}]}
    #x.update(d)

    from collections import namedtuple

    d_named = namedtuple('Struct', dict_a.keys())(*dict_a.values())
    print(d_named)

    """
